/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Parallel I/O functions
 *
 * $Id: io.c,v 1.8 2010/07/01 05:50:47 zlb Exp $
 */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>	/* PATH_MAX */

#if USE_MPI
static struct {		/* struct for sorting maps (L2Gmap_vert) */
    INT gindex, lindex;
} *map_sort, *ms0, *ms1;

static int
map_sort_compare(const void *m0, const void *m1)
{
    ms0 = (void *)m0; ms1 = (void *)m1;
    return (int)(ms0->gindex - ms1->gindex);
}

static int
merge_blocks(int n, int *blocklens, int *indices)
{
    int *blocklens0 = blocklens, *p = indices;

    if (n <= 0)
	return n;

    while (--n > 0) {
	if (p[1] - p[0] == 1) {
	    ++(*blocklens);
	    ++p;
	}
	else {
	    ++blocklens;
	    *(++indices) = *(++p);
	}
    }

    return blocklens - blocklens0 + 1;
}
#endif	/* USE_MPI */

#if USE_MPIIO	/*------------------------ Use MPI I/O ----------------------*/

/* Use MPI-2 parallel I/O functions */

static MPI_File fh = MPI_FILE_NULL;

BOOLEAN
phgIOOpen(GRID *g, const char *fn)
{
    MPI_File_open(g->comm, (char *)fn,
			MPI_MODE_CREATE | MPI_MODE_WRONLY, MPI_INFO_NULL, &fh);
    if (fh == MPI_FILE_NULL) {
	return FALSE;
    }
    MPI_File_set_atomicity(fh, 0);
    MPI_File_set_size(fh, 0);	/* truncate initial file size to 0 */
    MPI_File_set_view(fh, 0, MPI_BYTE, MPI_BYTE, "native", MPI_INFO_NULL);

    return TRUE;
}

int
phgIOPrintf(GRID *g, char *fmt, ...)
/* serial fprintf analog to an MPI file: at first rank 0 writes the data,
   then new File View set to end of file */
{
    va_list ap;
    char s[1024];
    int ret = 0;
    MPI_Offset offset;
    MPI_Status status;

    if (g->rank == 0) {
	va_start(ap, fmt);
	ret = vsprintf(s, fmt, ap);
	va_end(ap);
	MPI_File_write(fh, s, strlen(s), MPI_BYTE, &status);
	MPI_File_get_position(fh, &offset);
    }
    MPI_Bcast(&offset, sizeof(offset), MPI_BYTE, 0, MPI_COMM_WORLD);
#if 0
    MPI_File_sync(fh);
    MPI_Barrier(g->comm);
    MPI_File_sync(fh);
#endif
    MPI_File_seek(fh, offset, MPI_SEEK_SET);
    MPI_File_seek_shared(fh, offset, MPI_SEEK_SET);

    return ret;
}

void
phgIOWriteRoot(GRID *g, void *buf, size_t size, int count)
{
    MPI_Offset offset;
    MPI_Status status;

    if (g->rank == 0) {
        MPI_File_write(fh, buf, size * count, MPI_BYTE, &status);
        MPI_File_get_position(fh, &offset);
    }
    MPI_Bcast(&offset, sizeof(offset), MPI_BYTE, 0, MPI_COMM_WORLD);
#if 0
    MPI_File_sync(fh);
    MPI_Barrier(g->comm);
    MPI_File_sync(fh);
#endif
    MPI_File_seek(fh, offset, MPI_SEEK_SET);
    MPI_File_seek_shared(fh, offset, MPI_SEEK_SET);

    return;
}

void
phgIOWrite(GRID *g, void *buffer, size_t size, size_t count, size_t nglobal,
		BTYPE flags[], INT map[], BOOLEAN map_flag)
/* Parallel data write.
 *
 * Each process has 'count' entries of size 'size' stored in 'buffer[]',
 * 'flags[i]' indicates whether entry i is owned by current process,
 * 'map[i]' is the global id of entry i 
 *
 * If flags == NULL then all entries are owned by current process
 * If map == NULL then map[i] == i + (sum of count of prev procs) is assumed.
 * If map_flag == FALSE the flags and map are not used.
 */
{
    MPI_Datatype type, type0;
    MPI_Status status;
    MPI_Offset offset, disp;
    INT i, j, k;
    int *blocklens, *indices;

    MPI_File_get_position(fh, &offset);
    MPI_File_get_byte_offset(fh, offset, &disp);
    if (g->nprocs > 1 && map_flag) {
	map_sort = phgAlloc(count * sizeof(*map_sort));
	if (count > 0) {
	    /* sort map[] (blocks must have increasing offsets in filetype) */
	    for (i = 0; i < count; i++) {
		map_sort[i].gindex = map[i];
		map_sort[i].lindex = i;
	    }
	    qsort(map_sort, count, sizeof(*map_sort), map_sort_compare);
	}
	blocklens = phgAlloc(count * sizeof(*blocklens));
	indices = phgAlloc(count * sizeof(*indices));
	j = 0;
	for (i = 0; i < count; i++) {
	    k = map_sort[i].lindex;
	    if (flags != NULL && !(flags[k] & OWNER)) continue;
	    blocklens[j] = 1;
	    indices[j++] = map_sort[i].gindex;
	}
	phgDebug((3, "Before merging contiguous blocks: %d blocks\n", j));
	j = merge_blocks(j, blocklens, indices); /* merge contiguous blocks */
	phgDebug((3, "After merging contiguous blocks:  %d blocks\n", j));
	MPI_Type_contiguous(size, MPI_BYTE, &type0);
	MPI_Type_commit(&type0);
	MPI_Type_indexed(j, blocklens, indices, type0, &type);
	MPI_Type_commit(&type);
	MPI_File_set_view(fh, disp, type0, type, "native", MPI_INFO_NULL);
	MPI_Type_free(&type);
	j = 0;
	for (i = 0; i < count; i++) {
	    k = map_sort[i].lindex;
	    if (flags != NULL && !(flags[k] & OWNER)) continue;
	    blocklens[j] = 1;
	    indices[j++] = k;
	}
	j = merge_blocks(j, blocklens, indices); /* merge contiguous blocks */
	MPI_Type_indexed(j, blocklens, indices, type0, &type);
	MPI_Type_commit(&type);
	MPI_File_write_all(fh, buffer, 1, type, &status);
	MPI_Type_free(&type);
	MPI_Type_free(&type0);
	MPI_File_set_view(fh, disp + nglobal * size, MPI_BYTE, MPI_BYTE,
		"native", MPI_INFO_NULL);
	phgFree(blocklens);
	phgFree(indices);
	phgFree(map_sort);
    }
    else {
#if 0	/* FIXME: using write_ordered gets an "io No such file or directory"
	   warning when closing the file on non root process with MPICH 1.2.7
	   ch_p4 */
	MPI_Type_contiguous(count * size, MPI_BYTE, &type);
	MPI_Type_commit(&type);
	MPI_File_write_ordered(fh, buffer, 1, type, &status);
	MPI_File_seek(fh, offset + nglobal * size, MPI_SEEK_SET);
	MPI_File_seek_shared(fh, offset + nglobal * size, MPI_SEEK_SET);
	MPI_Type_free(&type);
#else
	j = count;
	MPI_Scan(&j, &k, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
	k -= j;		/* k = location */
	j = 1;		/* j = blocksize */
	MPI_Type_contiguous(size, MPI_BYTE, &type0);
	MPI_Type_commit(&type0);
	MPI_Type_indexed(1, &j, &k, type0, &type);
	MPI_Type_commit(&type);
	MPI_File_set_view(fh, disp, type0, type, "native", MPI_INFO_NULL);
	MPI_Type_free(&type);
	MPI_File_write_all(fh, buffer, count, type0, &status);
	MPI_Type_free(&type0);
	MPI_File_set_view(fh, disp + nglobal * size, MPI_BYTE, MPI_BYTE,
		"native", MPI_INFO_NULL);
#endif
    }
}

void
phgIOClose()
{
    MPI_File_close(&fh);
    fh = MPI_FILE_NULL;
}

#else	/*------------------------ Use sequaential I/O ---------------------*/

/* Use sequential I/O functions */

static FILE *fp = NULL;

BOOLEAN
phgIOOpen(GRID *g, const char *fn)
{
    if (g->rank > 0)
	return TRUE;
    return (fp = fopen(fn, "wb")) != NULL;
}

int
phgIOPrintf(GRID *g, char *fmt, ...)
/* serial fprintf analog to an MPI file: at first rank 0 writes the data,
   then new File View set to end of file */
{
    va_list ap;
    int ret;

    if (g->rank > 0)
	return 0; 

    va_start(ap, fmt);
    ret = vfprintf(fp, fmt, ap);
    va_end(ap);

    return ret;
}

void
phgIOWriteRoot(GRID *g, void *buf, size_t size, int count)
{
    if (g->rank > 0)
        return; 

    if (fwrite(buf, size, count, fp) != count)
	phgError(1, "%s: fwrite error, abort.\n", __func__);
    return;
}

void
phgIOWrite(GRID *g, void *buffer, size_t size, size_t count, size_t nglobal,
		BTYPE flags[], INT map[], BOOLEAN map_flag)
/* Parallel data write.
 *
 * Each process has 'count' entries of size 'size' stored in 'buffer[]',
 * 'flags[i]' indicates whether entry i is owned by current process,
 * 'map[i]' is the global id of entry i 
 *
 * If flags == NULL then all entries are owned by current process
 * If map == NULL then map[i] == i + (sum of count of prev procs) is assumed.
 */
{
#if USE_MPI
    if (g->nprocs > 1) {
	MPI_Datatype type, type0;
	MPI_Status status;
	MPI_Request req;
	INT i, j, k;
	int *blocklens, *indices, count_max, pass;
	long pos = 0;
	unsigned char *buf, *ptr;

	if (g->rank == 0)
	    pos = ftell(fp);

	pass = count;
	MPI_Allreduce(&pass, &count_max, 1, MPI_INT, MPI_MAX, g->comm);
	buf = phgAlloc(size * count_max);
	blocklens = phgAlloc(2 * (count_max + 1) * sizeof(*blocklens));
	indices = blocklens + count_max + 1;
	map_sort = phgAlloc(count * sizeof(*map_sort));

	j = 0;
	if (!map_flag) {
	    int tmp = count, map0;
	    MPI_Scan(&tmp, &map0, 1, MPI_INT, MPI_SUM, g->comm);
	    map0 -= tmp;
	    for (i = 0; i < count; i++) {
		map_sort[i].gindex = i + map0;
		map_sort[i].lindex = i;
	    }
	}
	else if (count > 0) {
	    /* sort map[] (blocks must have increasing offsets in filetype) */
	    for (i = 0; i < count; i++) {
		map_sort[i].gindex = map[i];
		map_sort[i].lindex = i;
	    }
	    qsort(map_sort, count, sizeof(*map_sort), map_sort_compare);
	}

	j = 0;
	for (i = 0; i < count; i++) {
	    k = map_sort[i].lindex;
	    if (flags != NULL && !(flags[k] & OWNER)) continue;
	    blocklens[j] = 1;
	    indices[j++] = k;
	}
	j = merge_blocks(j, blocklens, indices);
	MPI_Type_contiguous(size, MPI_BYTE, &type0);
	MPI_Type_commit(&type0);
	MPI_Type_indexed(j, blocklens, indices, type0, &type);
	MPI_Type_commit(&type);

	j = 0;
	for (i = 0; i < count; i++) {
	    k = map_sort[i].lindex;
	    if (flags != NULL && !(flags[k] & OWNER)) continue;
	    blocklens[j] = 1;
	    indices[j++] = map_sort[i].gindex;
	}
	count = merge_blocks(j, blocklens, indices);
	blocklens[count] = -1;
	indices[count] = j;

	for (pass = 0; pass < g->nprocs; pass++) {
	    if (g->rank != 0 && g->rank != pass)
		continue;

	    if (pass != 0) {
		/* send blocklens/indices to proc 0 */
		if (g->rank == pass)
		    MPI_Send(blocklens, 2*(count_max+1) * sizeof(*blocklens),
			 MPI_BYTE, 0, 11, g->comm);
		else
		    MPI_Recv(blocklens, 2*(count_max+1) * sizeof(*blocklens),
			 MPI_BYTE, pass, 11, g->comm, &status);
	    }

	    /* send data to process 0 */
	    if (g->rank == pass) {
		MPI_Isend(buffer, 1, type, 0, 22, g->comm, &req);
	    }
	    if (g->rank == 0) {
		/* get count */
		j = 0;
		while (blocklens[j] >= 0)
		    j++;
		MPI_Recv(buf, indices[j], type0, pass, 22, g->comm, &status);
	    }
	    if (g->rank == pass)
		MPI_Wait(&req, &status);

	    if (g->rank != 0)
		continue;

	    /* write data */
	    ptr = buf;
	    for (j = 0; blocklens[j] >= 0; j++) {
		fseek(fp, pos + indices[j] * size, SEEK_SET);
		if (fwrite(ptr, size, blocklens[j], fp) != blocklens[j])
		    phgError(1, "%s: fwrite error, abort.\n", __func__);
		ptr += blocklens[j] * size;
	    }
	}	/* for (pass ...) */

	MPI_Type_free(&type);
	MPI_Type_free(&type0);

	phgFree(blocklens);
	phgFree(map_sort);
	phgFree(buf);

	if (g->rank == 0)
	    fseek(fp, pos + nglobal * size, SEEK_SET);

	return;
    }
#endif	/* USE_MPI */

    assert(count == nglobal);
    if (fwrite(buffer, size, count, fp) != count)
	phgError(1, "%s: fwrite error, abort.\n", __func__);;
}

void
phgIOClose(void)
{
    if (fp != NULL) {
	fclose(fp);
	fp = NULL;
    }
}

#endif	/*----------------------------- USE_MPIIO --------------------------*/

const char *
phgIOAddExtension(const char *fn, const char *ext)
/* adds '.ext' to the given filename if the latter does not contain
 * an extension, returns the address of the new filename which points either
 * to 'fn' or a static buffer */
{
    static char new_name[PATH_MAX + 1];
    const char *p;

    if (ext == NULL)
	return fn;
    if ((p = strrchr(fn, '/')) == NULL)
	p = fn;
    if ((p = strchr(p, '.')) != NULL)
	return fn;
    sprintf(new_name, "%s.%s", fn, ext);
    return new_name;
}
